import traceback
from logger import Logger


class PrintLogger(Logger):
    def __init__(self):
        super(PrintLogger, self).__init__('PrintLogger')

    def _log(self, priority, message):
        print(message)

    def traceback(self):
        print('')
        print(traceback.format_exc())



import sys
from setting import Setting


def main():
    """
    Main function
    """
    setting = Setting().parse_args(sys.argv)
    setting.logger.info('Script started: args=%s' % sys.argv[1:])

    try:
        setting = setting.load_config()
        for watcher in setting.watchers:
            for alert in watcher.watch():
                for notifier in setting.notifiers:
                    notifier.notify(alert)
            watcher.after_success()
    except Exception as e:
        setting.logger.error('Script ended with error: %s: %s' % (e.__class__.__name__, e))
        setting.logger.traceback()
        return 1

    setting.logger.info('Script ended successfully.')
    return 0



from abc import ABCMeta, abstractmethod
from easy_alert.util import CaseClass


class Watcher(CaseClass):
    __metaclass__ = ABCMeta

    def __init__(self, **args):
        super(Watcher, self).__init__(args.keys())
        for k in args:
            setattr(self, k, args[k])

    @abstractmethod
    def watch(self):
        """abstract method"""

    def after_success(self):
        """do nothing in default"""



import logging
import paramiko
import subprocess
import re
import itertools

from datetime import datetime
from watcher import Watcher
from easy_alert.entity import Alert, Level
from easy_alert.util import get_server_id
from easy_alert.i18n import *
from easy_alert.setting.setting_error import SettingError


class SSHWatcher(Watcher):
    """
    Check SSH connection to servers

    configuration should be a list of this dict
      user [required]: user name to log in
      key  [required]: path to the SSH private key
      port           : SSH port number (default:22)
      dynamic        : command line string for getting dynamic host definitions
                       The output should be '<host> <name>' for each line.
      host        [*]: host name or IP address of the host to check connection [*required if not dynamic]
      name        [*]: short description of the host [*required if not dynamic]

    The alert level is always ERR.
    """
    DEFAULT_PORT = 22
    CONNECTION_TIMEOUT = 10  # seconds

    def __init__(self, ssh_setting):
        if not isinstance(ssh_setting, list):
            raise SettingError('SSHWatcher settings not a list: %s' % ssh_setting)

        settings = []
        for s in ssh_setting:
            if not isinstance(s, dict):
                raise SettingError('SSHWatcher settings not a dict: %s' % s)

            try:
                user = s['user']
                key = s['key']
                port = int(s.get('port', self.DEFAULT_PORT))
                dynamic_cmd = s.get('dynamic')
                is_dynamic = dynamic_cmd is not None
                name = None if is_dynamic else s['name']
                host = None if is_dynamic else s['host']
                settings.append((user, key, port, is_dynamic, name, host, dynamic_cmd))
            except KeyError as e:
                raise SettingError('SSHWatcher not found config key: %s' % e)
            except Exception as e:
                raise SettingError('SSHWatcher settings syntax error: %s' % e)

        super(SSHWatcher, self).__init__(settings=settings)

    @staticmethod
    def _build_target_list(s):
        """
        :param s: tuple (user, key, port, is_dynamic, name, host, dynamic_cmd)
        :return: list of the tuple (name, host, port, user, key)
        """
        user, key, port, is_dynamic, name, host, dynamic_cmd = s

        if not is_dynamic:
            return [(name, host, port, user, key)]

        ret = []
        p = subprocess.Popen(dynamic_cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        for line in p.stdout:
            r = re.compile('^\s*(.+)\s+(.+)\s*$')
            m = r.match(line)
            host, name = m.group(1, 2)
            ret.append((name, host, port, user, key))
        return ret

    def watch(self):
        """
        :return: list of Alert instances
        """
        start_time = datetime.now()

        target = list(itertools.chain.from_iterable(self._build_target_list(s) for s in self.settings))

        errors = []
        for name, host, port, user, key in target:
            is_success, msg = self._check_ssh_connection(host, port, user, key)
            if not is_success:
                m = MSG_SSH_STATUS_FORMAT % {'name': name, 'user': user, 'host': host, 'port': port, 'msg': msg}
                errors.append(m)

        if errors:
            message = MSG_SSH_ALERT % {'server_id': get_server_id(), 'result': '\n'.join('%s' % s for s in errors)}
            return [Alert(start_time, Level(logging.ERROR), MSG_SSH_ALERT_TITLE, message)]
        else:
            return []

    @staticmethod
    def _check_ssh_connection(host, port, user, key):
        """
        :return: (is success, error message)
        """
        client = paramiko.SSHClient()
        client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

        try:
            client.connect(host, port, user, key_filename=key, timeout=10)
            return True, None
        except Exception as e:
            return False, '%s: %s' % (e.__class__.__name__, str(e))
        finally:
            client.close()



import re
import urllib2
from urlparse import urlparse
from datetime import datetime
from watcher import Watcher
from easy_alert.entity import Alert, Level
from easy_alert.i18n import *
from easy_alert.setting.setting_error import SettingError
from easy_alert.util import Matcher, get_server_id, apply_option, with_retry


class HTTPWatcher(Watcher):
    """
    Check http(s) connection and response to the url

    configuration should be a list of this dict
      name    [required]: short description for the url
      level   [required]: alert level (should be in {critical, error, warn, info, debug})
      url     [required]: url string (scheme should be http or https)
      timeout           : connection timeout in second (default:10)
      retry             : max count to retry (default:0)
      additional_info   : additional information string (default:empty string)
      expect_code    [*]: expected status code of the response
      expect_size    [*]: expected data size in matcher format (e.g. '=123', '< 123', '>=123')
      expect_regexp  [*]: expected content text in regexp

    * Any of {expect_code, expect_size, expect_text} is required.
    """
    DEFAULT_TIMEOUT = 10

    def __init__(self, http_setting):
        if not isinstance(http_setting, list):
            raise SettingError('HTTPWatcher settings not a list: %s' % http_setting)

        settings = []
        for s in http_setting:
            if not isinstance(s, dict):
                raise SettingError('HTTPWatcher settings not a dict: %s' % s)

            try:
                name = s['name']
                level = self.__get_level(s['level'])
                url = s['url']
                timeout = int(s.get('timeout', self.DEFAULT_TIMEOUT))
                retry = int(s.get('retry', 0))
                additional_info = s.get('additional_info', '')
                expect_code = apply_option(int, s.get('expect_code'))
                expect_size = apply_option(Matcher, s.get('expect_size'))
                expect_regexp = apply_option(re.compile, s.get('expect_regexp'))
            except SettingError as e:
                raise e
            except KeyError as e:
                raise SettingError('HTTPWatcher not found config key: %s' % e)
            except Exception as e:
                raise SettingError('HTTPWatcher settings syntax error: %s' % e)

            self.__verify_url(url)
            if all([expect_code is None, expect_size is None, expect_regexp is None]):
                raise SettingError('HTTPWatcher any of expect_code, expect_size or expect_regexp should be set.')

            settings.append((
                name, level, url, timeout, retry, additional_info, expect_code, expect_size, expect_regexp))

        super(HTTPWatcher, self).__init__(settings=settings)

    @staticmethod
    def __verify_url(url):
        parsed_url = urlparse(url)
        scheme = parsed_url.scheme or 'http'
        if scheme not in ['http', 'https']:
            raise SettingError('HTTPWatcher unsupported scheme: %s' % scheme)

    @staticmethod
    def __get_level(level_str):
        level = [l for l in Level.seq if l.get_keyword() == level_str]
        if not level:
            raise SettingError('HTTPWatcher invalid level: %s' % level_str)
        return level[0]

    @staticmethod
    def _connect_url(url, timeout):
        u = urllib2.urlopen(url, timeout=timeout)
        data = u.read()
        return u.getcode(), data

    @staticmethod
    def _should_alert(code, data, expect_code, expect_size, expect_regexp):
        if expect_code is not None:
            if code != expect_code:
                return True
        if expect_size is not None:
            if not expect_size.check(len(data)):
                return True
        if expect_regexp is not None:
            if not expect_regexp.findall(data):
                return True
        return False

    def watch(self):
        """
        :return: list of Alert instances
        """
        start_time = datetime.now()

        result = []
        for name, level, url, timeout, retry, additional_info, expect_code, expect_size, expect_regexp in self.settings:
            code, data = with_retry(retry)(lambda: self._connect_url(url, timeout))
            if self._should_alert(code, data, expect_code, expect_size, expect_regexp):
                message = MSG_HTTP_ALERT_FORMAT % {
                    'level': level.get_text(),
                    'name': name,
                    'url': url,
                    'code': code,
                    'size': len(data),
                    'expect_code': expect_code,
                    'expect_size': expect_size,
                    'expect_regexp': expect_regexp.pattern if expect_regexp else None,
                    'additional_info': additional_info,
                }
                result.append((level, message))

        if result:
            max_level = max(r[0] for r in result)
            message = MSG_HTTP_ALERT % {'server_id': get_server_id(), 'result': '\n'.join(r[1] for r in result)}
            return [Alert(start_time, max_level, MSG_HTTP_ALERT_TITLE, message)]
        else:
            return []



from optparse import OptionParser
from default import VERSION, USAGE, DEFAULT_CONF_PATH


def get_parser():
    """
    Get command line arguments parser
    """
    parser = OptionParser(version=VERSION, usage=USAGE)

    parser.add_option(
        '--config', dest='config_path', default=DEFAULT_CONF_PATH, type='string',
        help='path to the config file'
    )
    parser.add_option(
        '--check', action='store_true', dest='print_only', default=False,
        help='print process information instead of sending notifications'
    )

    return parser



import re
from case_class import CaseClass


class Matcher(CaseClass):
    def __init__(self, s):
        super(Matcher, self).__init__(['operator', 'threshold'])

        m = re.match("""([=!<>]+)\s*(\d+)""", s)
        if m is None:
            raise Exception('Invalid matcher string: %s' % s)

        op, tt = m.groups()
        t = int(tt)

        if op == '==':
            op = '='
        f = {
            '=': lambda x: x == int(t),
            '!=': lambda x: x != int(t),
            '<': lambda x: x < int(t),
            '<=': lambda x: x <= int(t),
            '>': lambda x: x > int(t),
            '>=': lambda x: x >= int(t),
        }.get(op)

        if f is None:
            raise Exception('Invalid matcher string: %s' % s)
        self.operator = op
        self.threshold = t
        self.check = f

    def __str__(self):
        return '"%s %s"' % (self.operator, self.threshold)



import arg_parser
import sys
import yaml
from easy_alert.logger import SystemLogger, PrintLogger
from easy_alert.util import CaseClass
from default import WATCHER_KEYS, WATCHER_FACTORIES, NOTIFIER_FACTORIES
from setting_error import SettingError


class Setting(CaseClass):
    """
    Manages all settings
    """

    # argument parser
    parser = arg_parser.get_parser()

    def __init__(self, watcher_types=list(), config_path=None, print_only=None, watchers=list(), notifiers=list(),
                 logger=SystemLogger()):
        super(Setting, self).__init__([
            'watcher_types', 'config_path', 'print_only', 'watchers', 'notifiers', 'logger',
        ])
        self.watcher_types = watcher_types
        self.config_path = config_path
        self.print_only = print_only
        self.watchers = watchers
        self.notifiers = notifiers
        self.logger = logger

    def parse_args(self, argv):
        """Parse command line arguments and update operation and options"""

        options, args = Setting.parser.parse_args(argv[1:])

        # args should be at least one and all watcher should be known
        if not args or set(args) - set(WATCHER_KEYS):
            self.parser.print_usage(sys.stderr)
            self.parser.exit(2)

        return Setting(sorted(set(args), key=args.index), options.config_path, options.print_only, self.watchers,
                       self.notifiers, PrintLogger() if options.print_only else self.logger)

    def load_config(self):
        """Load configuration file"""

        # assertions
        assert self.watcher_types, 'watcher_types should be set before load_config'
        assert (self.config_path is not None), 'config_path should be set before load_config'
        assert (self.print_only is not None), 'print_only should be set before load_config'

        config = yaml.load(open(self.config_path).read().decode('utf-8'))
        if not isinstance(config, dict):
            raise SettingError('Syntax error: %s' % self.config_path)

        # parse watchers
        ws = config.get('watchers')
        if ws:
            if not isinstance(ws, dict):
                raise SettingError('Syntax error: %s' % self.config_path)
            watchers = [self._parse_watcher(t, ws) for t in self.watcher_types]
        else:
            raise SettingError('Not found "watchers" entry: %s' % self.config_path)

        # parse notifiers
        ns = config.get('notifiers')
        if ns:
            if not isinstance(ns, dict):
                raise SettingError('Syntax error: %s' % self.config_path)
            assert ns  # it should not be empty
            notifiers = [self._parse_notifier(k, v) for k, v in ns.items()]
        else:
            raise SettingError('Not found "notifiers" entry: %s' % self.config_path)

        return Setting(self.watcher_types, self.config_path, self.print_only, watchers, notifiers, self.logger)

    def _parse_watcher(self, watcher_type, watcher_config):
        factory = WATCHER_FACTORIES.get(watcher_type)
        if not factory:
            raise SettingError('Unsupported watcher type: %s' % watcher_type)

        conf = watcher_config.get(watcher_type)
        if not conf:
            raise SettingError('Not found watcher configuration for "%s": %s' % (watcher_type, self.config_path))

        return factory(conf, self.print_only, self.logger)

    def _parse_notifier(self, notifier_type, notifier_config):
        factory = NOTIFIER_FACTORIES.get(notifier_type)
        if not factory:
            raise SettingError('Unsupported notifier type: %s in %s' % (notifier_type, self.config_path))

        return factory(notifier_config, self.print_only, self.logger)



import os
import locale


def __get_locale():
    # environment LANG is the first priority
    lang = os.environ.get('LANG')
    if lang:
        return lang.lower()
    return locale.getdefaultlocale()[0].lower()


__locale = __get_locale()

if __locale.startswith('ja_'):
    from messages_ja import *
else:
    from messages_en import *



from logging import DEBUG, INFO, WARN, ERROR, CRITICAL
from easy_alert.util import CaseClass
from easy_alert.i18n import *


class Level(CaseClass):
    def __init__(self, level):
        super(Level, self).__init__(['level'])
        self.level = level

    def get_text(self):
        mapping = {CRITICAL: MSG_CRITICAL, ERROR: MSG_ERROR, WARN: MSG_WARN, INFO: MSG_INFO, DEBUG: MSG_DEBUG}
        return mapping.get(self.level)

    def get_keyword(self):
        mapping = {CRITICAL: 'critical', ERROR: 'error', WARN: 'warn', INFO: 'info', DEBUG: 'debug'}
        return mapping[self.level]

    seq = None


Level.seq = [Level(l) for l in [CRITICAL, ERROR, WARN, INFO, DEBUG]]



class CaseClass(object):
    """
    Implementation like Scala's case class
    """

    def __init__(self, keys):
        """
        :param keys: list of attribute names
        """
        self.__keys = keys

    def __cmp__(self, other):
        if not isinstance(other, self.__class__):
            # compare with class names
            return cmp(self.__class__.__name__, other.__class__.__name__)

        for k in self.__keys:
            a, b = getattr(self, k), getattr(other, k)
            if a < b:
                return -1
            if a > b:
                return 1
        return 0

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, ', '.join('%s=%r' % (k, getattr(self, k)) for k in self.__keys))



from easy_alert.watcher import ProcessWatcher, LogWatcher, SSHWatcher, CommandWatcher, HTTPWatcher
from easy_alert.notifier import EmailNotifier

# dict of the watcher and its factory
WATCHER_FACTORIES = {
    'log': lambda conf, print_only, logger: LogWatcher(conf, print_only),
    'process': lambda conf, print_only, logger: ProcessWatcher(conf),
    'ssh': lambda conf, print_only, logger: SSHWatcher(conf),
    'command': lambda conf, print_only, logger: CommandWatcher(conf),
    'http': lambda conf, print_only, logger: HTTPWatcher(conf),
}

# dict of the notifier and its factory
NOTIFIER_FACTORIES = {
    'email': lambda conf, print_only, logger: EmailNotifier(conf, print_only, logger),
}

WATCHER_KEYS = sorted(WATCHER_FACTORIES.keys())

VERSION = 'easy-alert %s' % __import__('easy_alert').__version__

USAGE = """
%prog [options] <watcher_type> [<watcher_type> ...]

Available watcher types:
  """ + '\n  '.join(WATCHER_KEYS)

DEFAULT_CONF_PATH = '/etc/easy-alert/easy-alert.yml'



def apply_option(f, arg):
    """
    Apply function if the arg is not None
    :param f: function to apply
    :param arg: argument value
    :return: None if the arg is None, else the result of the function
    """
    return arg if arg is None else f(arg)


def with_retry(count):
    """
    Retryable function execution
    :param count: retry limit
    :return:
    """
    def f(thunk, c=count):
        while True:
            try:
                return thunk()
            except Exception as e:
                if c > 0:
                    c -= 1
                else:
                    raise e
    return f



from easy_alert.util import CaseClass


class Alert(CaseClass):
    """
    Abstract model which can be the output of watchers and the input of notifiers
    """
    keys = ['start_time', 'level', 'title', 'message']

    def __init__(self, start_time, level, title, message):
        super(Alert, self).__init__(self.keys)
        self.start_time = start_time
        self.level = level
        self.title = title
        self.message = message



import os
import sys
import syslog
from logger import Logger


class SystemLogger(Logger):
    def __init__(self):
        self.name = os.path.basename(sys.argv[0])
        super(SystemLogger, self).__init__('SystemLogger[%s]' % self.name)

    def _log(self, priority, message):
        syslog.openlog(self.name, syslog.LOG_PID)
        syslog.syslog(priority, message)
        syslog.closelog()



from abc import ABCMeta, abstractmethod
from easy_alert.util import CaseClass, get_server_id
from easy_alert.i18n import *


class Notifier(CaseClass):
    __metaclass__ = ABCMeta

    def __init__(self, group_id=None, **args):
        super(Notifier, self).__init__(['group_id'] + args.keys())
        self.group_id = group_id
        for k in args:
            setattr(self, k, args[k])

    def get_subject(self, alert):
        return MSG_SUBJECT_FORMAT % {
            'level': alert.level.get_text(),
            'group_id': self.group_id,
            'server_id': get_server_id(),
            'title': alert.title,
            'start_time': alert.start_time.strftime('%Y-%m-%d %H:%M:%S')
        }

    @abstractmethod
    def notify(self, alert):
        """abstract method"""



import os
import glob
import json
from datetime import datetime
from collections import defaultdict
from logging import WARN, ERROR

from watcher import Watcher
from easy_alert.entity import Alert, Level
from easy_alert.util import get_server_id
from easy_alert.i18n import *
from easy_alert.setting.setting_error import SettingError


class LogWatcher(Watcher):
    """
    Watch the log files filtered by Fluentd (td-agent)

    configuration should be this dict
      watch_dir  [required]: path to the directory to watch
      target_pattern       : glob pattern for the target files
      pending_pattern      : glob pattern for checking pending files
      message_num_threshold: maximum number of messages for each log level
      message_len_threshold: maximum length for each message
      pending_threshold    : threshold number of files for checking pending files

    The settings of td-agent.conf should be like this.

    # Watch the log file
    <source>
      type tail
      path /var/log/messages
      pos_file /var/log/td-agent/tail.syslog.pos
      tag monitor.syslog
      format none
    </source>

    # Rewrite the tag for each line with its severity
    <match monitor.syslog>
      type rewrite_tag_filter
      rewriterule100 message ERROR ${tag}.error
      rewriterule500 message WARN ${tag}.warn
      rewriterule999 message .* clear
    </match>

    # Write to the temporary alert queue
    <match monitor.*.*>
      type file
      path /var/log/easy-alert/alert
      time_slice_format %Y%m%d_%H%M
      time_slice_wait 5s
      time_format %Y%m%dT%H%M%S%z
      localtime
      buffer_chunk_limit 256m
    </match>

    # Logs to be ignored
    <match clear>
      type null
    </match>
    """

    DEFAULT_TARGET_PATTERN = 'alert.????????_????_*.log'
    DEFAULT_PENDING_PATTERN = 'alert.????????_????*'
    DEFAULT_MESSAGE_NUM_THRESHOLD = 15
    DEFAULT_MESSAGE_LEN_THRESHOLD = 1024
    DEFAULT_PENDING_THRESHOLD = 3

    def __init__(self, alert_setting, print_only):
        if not isinstance(alert_setting, dict):
            raise SettingError('LogWatcher settings not a dict: %s' % alert_setting)

        try:
            watch_dir = alert_setting['watch_dir']
            tp = os.path.join(watch_dir, alert_setting.get('target_pattern') or self.DEFAULT_TARGET_PATTERN)
            pp = os.path.join(watch_dir, alert_setting.get('pending_pattern') or self.DEFAULT_PENDING_PATTERN)
            nt = int(alert_setting.get('message_num_threshold') or self.DEFAULT_MESSAGE_NUM_THRESHOLD)
            lt = int(alert_setting.get('message_len_threshold') or self.DEFAULT_MESSAGE_LEN_THRESHOLD)
            pt = int(alert_setting.get('pending_threshold') or self.DEFAULT_PENDING_THRESHOLD)
        except KeyError as e:
            raise SettingError('LogWatcher not found config key: %s' % e)
        except Exception as e:
            raise SettingError('LogWatcher settings syntax error: %s' % e)

        super(LogWatcher, self).__init__(
            watch_dir=watch_dir, target_pattern=tp, pending_pattern=pp, message_num_threshold=nt,
            message_len_threshold=lt, pending_threshold=pt, print_only=print_only, target_paths=None
        )

    def watch(self):
        """
        :return: list of Alert instances
        """
        start_time = datetime.now()

        # get target paths
        self.target_paths = glob.glob(self.target_pattern)
        if not self.target_paths:
            return self._check_pending(start_time)

        # parse files
        result, max_level = self._parse_files(self.target_paths)

        # make alert
        message = MSG_LOG_ALERT % {'server_id': get_server_id(), 'result': self._make_result(result)}
        return [Alert(start_time, max_level, MSG_LOG_ALERT_TITLE, message)]

    def after_success(self):
        """Delete parsed files after the notification."""

        assert (self.target_paths is not None)

        for path in self.target_paths:
            if self.print_only:
                print('Would remove: %s' % path)
            else:
                os.remove(path)

    def _check_pending(self, start_time):
        paths = glob.glob(self.pending_pattern)
        ret = []
        if len(paths) >= self.pending_threshold:
            mapping = {'server_id': get_server_id(), 'pattern': self.pending_pattern, 'paths': '\n'.join(paths)}
            ret.append(Alert(start_time, Level(WARN), MSG_LOG_ALERT_PENDING_TITLE, MSG_LOG_ALERT_PENDING % mapping))
        return ret

    def _parse_files(self, paths):
        d = defaultdict(lambda: (0, []))
        max_level = Level(WARN)
        for path in paths:
            for line in open(path):
                tokens = line.split('\t')
                tag = tokens[1]
                msg = json.loads(tokens[2].decode('utf-8', 'ignore'))['message']
                cnt, msgs = d[tag]
                cnt += 1
                # trim string
                msgs += [msg[:self.message_len_threshold]] if cnt <= self.message_num_threshold else []
                d[tag] = (cnt, msgs)
                if max_level == Level(WARN) and tag.endswith('.error'):
                    max_level = Level(ERROR)
        return d, max_level

    def _make_result(self, result):
        buf = []
        for k in sorted(result.keys()):
            cnt, msgs = result[k]
            buf.append(MSG_LOG_SUMMARY % {'tag': k, 'count': cnt})
            buf += msgs
            buf += [MSG_LOG_SNIP] if cnt > self.message_num_threshold else []
            buf.append('')
        return '\n'.join(buf)



# -*- coding: utf-8 -*-

EMAIL_ENCODING = 'iso-2022-jp'

MSG_DEBUG = u"デバッグ"
MSG_INFO = u"情報"
MSG_WARN = u"警戒"
MSG_ERROR = u"危険"
MSG_CRITICAL = u"緊急"

MSG_PROC_NOT_RUNNING = u'起動していません'
MSG_PROC_RUNNING = u'%(count)d 個 起動しています'
MSG_PROC_STATUS_FORMAT = u'[%(level)s] %(name)s: %(count)s ("%(condition)s" に違反)'
MSG_PROC_ALERT_TITLE = u'プロセス異常を検知しました'
MSG_PROC_ALERT = u'サーバ %(server_id)s にて、以下のプロセス異常を検知しました。\n\n%(result)s\n\n以上'

MSG_LOG_SUMMARY = u'[%(tag)s]: %(count)d件'
MSG_LOG_SNIP = u'(省略されました)'
MSG_LOG_ALERT_TITLE = u'エラーログを検知しました'
MSG_LOG_ALERT = u'サーバ %(server_id)s にて、以下のエラーを検知しました。\n\n%(result)s\n以上'
MSG_LOG_ALERT_PENDING_TITLE = u'ログ監視が滞留している可能性があります'
MSG_LOG_ALERT_PENDING = u"""サーバ %(server_id)s にて、ログ監視滞留の兆候を検知しました。\n
パターン %(pattern)s にマッチするファイルが複数存在します。\n\n%(paths)s\n
ファイルの状態を確認し、問題があれば Fluentd および各アプリケーションの再起動を試みてください。"""

MSG_SSH_STATUS_FORMAT = u'[%(name)s](%(user)s@%(host)s:%(port)d): %(msg)s'
MSG_SSH_ALERT_TITLE = u'SSH疎通異常を検知しました'
MSG_SSH_ALERT = u'サーバ %(server_id)s から、以下のサーバに対する SSH 疎通確認に失敗しました。\n\n%(result)s\n\n以上'

MSG_CMD_ALERT_FORMAT = u"""[%(level)s] %(name)s のヘルスチェックに失敗
  actual: {code:%(code)d, stdout:%(stdout)s, stderr:%(stderr)s}
  expect: {code:%(expect_code)d, stdout:%(expect_stdout)s, stderr:%(expect_stderr)s}"""
MSG_CMD_ALERT_TITLE = u'ヘルスチェック異常を検知しました'
MSG_CMD_ALERT = u'サーバ %(server_id)s にて、以下のヘルスチェック異常を検知しました。\n\n%(result)s\n\n以上'

MSG_HTTP_ALERT_FORMAT = u"""[%(level)s] %(name)s のヘルスチェックに失敗
  url    : %(url)s
  actual : {code:%(code)d, size:%(size)d}
  expect : {code:%(expect_code)d, size:%(expect_size)s, regexp:%(expect_regexp)s}
  message: %(additional_info)s"""
MSG_HTTP_ALERT_TITLE = u'HTTP疎通異常を検知しました'
MSG_HTTP_ALERT = u'サーバ %(server_id)s にて、以下のヘルスチェック異常を検知しました。\n\n%(result)s\n\n以上'

MSG_SUBJECT_FORMAT = u'【%(level)s】[%(group_id)s:%(server_id)s] %(title)s (%(start_time)s)'



import re
from collections import defaultdict
import subprocess
from datetime import datetime

from watcher import Watcher
from easy_alert.entity import Alert, Level
from easy_alert.util import CaseClass, get_server_id
from easy_alert.i18n import *
from easy_alert.setting.setting_error import SettingError


class ProcessReader(object):
    """
    Read process information from operating system
    """

    def read(self):
        """
        Get the running processes information
        :return: dict of process id -> tuple(parent process id, args string)
        """

        def f(line):
            tokens = line.split(None, 2)
            return int(tokens[0]), (int(tokens[1]), tokens[2])

        # trim header line then parse
        return dict(map(f, self._read_raw().splitlines()[1:]))

    def _read_raw(self):
        cmd = ['/bin/ps', 'ax', '-o', 'pid,ppid,args']
        return subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE).communicate()[0]


class ProcessCounter(CaseClass):
    """
    Count the number of the processes
    """

    def __init__(self, process_dict):
        super(ProcessCounter, self).__init__(['process_dict', 'cache_distinct', 'cache_aggregated'])
        self.process_dict = process_dict
        self.cache_distinct = None
        self.cache_aggregated = None

    def count(self, pattern, aggregate=True):
        d = self._aggregate() if aggregate else self._distinct()
        return sum(x * bool(pattern.search(s)) for s, x in d.items())

    def _distinct(self):
        """
        :return: dict of args string -> count with default value (0)
        """
        if self.cache_distinct is None:
            d = defaultdict(int)
            for ppid, args in self.process_dict.values():
                d[args] += 1
            self.cache_distinct = d
        return self.cache_distinct

    def _aggregate(self):
        """
        Aggregate forked processes.
        That is, if the 'args' strings of one process is same as the parent process, it is not counted.

        :return: dict of args string -> count with default value (0)
        """
        if self.cache_aggregated is None:
            d = defaultdict(int)
            for ppid, args in self.process_dict.values():
                parent = self.process_dict.get(ppid)
                if parent is None or parent[1] != args:
                    d[args] += 1
            self.cache_aggregated = d

        return self.cache_aggregated


class Condition(CaseClass):
    """
    Manages comparison operator and the threshold.
    """

    def __init__(self, condition):
        super(Condition, self).__init__(['operator', 'threshold'])

        op, threshold = re.match("""([=!<>]+)\s*(\d+)""", condition).groups()
        t = int(threshold)
        f = {
            '=': lambda x: x == int(t),
            '==': lambda x: x == int(t),
            '!=': lambda x: x != int(t),
            '<': lambda x: x < int(t),
            '<=': lambda x: x <= int(t),
            '>': lambda x: x > int(t),
            '>=': lambda x: x >= int(t),
        }.get(op)

        if f is None:
            raise SettingError('ProcessWatcher invalid operator: %s' % op)
        self.operator = op
        self.threshold = int(threshold)
        self.check = f

    def __str__(self):
        return '%s %d' % (self.operator, self.threshold)


class ProcessWatcher(Watcher):
    """
    Watch the number of the running processes

    configuration should be a list of this dict
      name   [required]: short description for the process
      regexp [required]: process argument string to be counted in regular expression
      aggregate        : aggregate forked processes (counted as one) if true (default:true)
      critical         : check the threshold then alert with critical level
      error            : check the threshold then alert with error level
      warn             : check the threshold then alert with warn level
      info             : check the threshold then alert with info level
      debug            : check the threshold then alert with debug level

    Any of {critical, error, warn, info, debug} is required.
    """

    def __init__(self, alert_setting, process_reader=ProcessReader()):
        if not isinstance(alert_setting, list):
            raise SettingError('ProcessWatcher settings not a list: %s' % alert_setting)

        settings = []
        for s in alert_setting:
            if not isinstance(s, dict):
                raise SettingError('ProcessWatcher settings not a dict: %s' % s)

            try:
                name = s['name']
                pattern = re.compile(s['regexp'])
                aggregate = s.get('aggregate', True)
                conditions = self._parse_conditions(s)
            except KeyError as e:
                raise SettingError('ProcessWatcher not found config key: %s' % e)
            except Exception as e:
                raise SettingError('ProcessWatcher settings syntax error: %s' % e)

            settings.append((name, pattern, aggregate, conditions))

        super(ProcessWatcher, self).__init__(settings=settings, process_reader=process_reader)

    @staticmethod
    def _parse_conditions(setting):
        """
        :return: sorted list of tuple(Level, Condition) by level (severest first)
        """
        ret = []
        for l in Level.seq:
            v = setting.get(l.get_keyword())
            if v:
                ret.append((l, Condition(v)))
        if not ret:
            raise SettingError('ProcessWatcher not found threshold: %s' % setting)
        return ret

    def watch(self):
        """
        :return: list of Alert instances
        """
        start_time = datetime.now()
        pc = ProcessCounter(self.process_reader.read())

        result = []
        for name, pattern, aggregate, conditions in self.settings:
            st = self.ProcessStatus(name, pc.count(pattern, aggregate), conditions)
            if st.level:
                result.append(st)

        if result:
            max_level = max(r.level for r in result)
            message = MSG_PROC_ALERT % {'server_id': get_server_id(), 'result': '\n'.join('%s' % s for s in result)}
            return [Alert(start_time, max_level, MSG_PROC_ALERT_TITLE, message)]
        else:
            return []

    class ProcessStatus(CaseClass):
        def __init__(self, name, count, conditions):
            super(ProcessWatcher.ProcessStatus, self).__init__(['name', 'count', 'level', 'condition'])
            self.name = name
            self.count = count
            self.level = None
            self.condition = None

            # check condition
            for level, condition in conditions:
                # find the first False from the severest level
                if not condition.check(count):
                    self.level = level
                    self.condition = condition
                    break

        def __str__(self):
            count_msg = MSG_PROC_NOT_RUNNING if self.count == 0 else MSG_PROC_RUNNING % {'count': self.count}
            return MSG_PROC_STATUS_FORMAT % {
                'level': self.level.get_text(),
                'name': self.name,
                'count': count_msg,
                'condition': self.condition
            }



from setuptools import setup, find_packages

SRC_DIR = 'src'


def get_version():
    import sys

    sys.path[:0] = [SRC_DIR]
    return __import__('easy_alert').__version__


setup(
    name='easy-alert',
    version=get_version(),
    description='Super Simple Process Monitoring Tool',
    author='mogproject',
    author_email='mogproj@gmail.com',
    url='https://github.com/mogproject/easy-alert',
    install_requires=[
        'pyyaml',
        'paramiko',
    ],
    tests_require=[
        'unittest2',
    ],
    package_dir={'': SRC_DIR},
    packages=find_packages(SRC_DIR),
    include_package_data=True,
    test_suite='tests',
    entry_points="""
    [console_scripts]
    easy-alert = easy_alert.easy_alert:main
    """,
)



import subprocess
import re
from datetime import datetime
from watcher import Watcher
from easy_alert.entity import Alert, Level
from easy_alert.i18n import *
from easy_alert.setting.setting_error import SettingError
from easy_alert.util import get_server_id


class CommandWatcher(Watcher):
    """
    Check the result of executing the external command

    configuration should be a list of this dict
      name    [required]: short description for the command
      level   [required]: alert level (should be in {critical, error, warn, info, debug})
      command [required]: command line string
      expect_code       : expected exit code number
      expect_stdout  [*]: expected stdout pattern in regexp
      expect_stderr  [*]: expected stderr pattern in regexp
      max_output_len [*]: maximum length for each stdout/stderr

    * Any of {expect_code, expect_stdout, expect_stderr} is required.
    """

    DEFAULT_MAX_OUTPUT_LEN = 1024

    def __init__(self, command_setting):
        if not isinstance(command_setting, list):
            raise SettingError('CommandWatcher settings not a list: %s' % command_setting)

        settings = []
        for s in command_setting:
            if not isinstance(s, dict):
                raise SettingError('CommandWatcher settings not a dict: %s' % s)

            try:
                name = s['name']
                level_str = s['level']
                level = [l for l in Level.seq if l.get_keyword() == level_str]
                command = s['command']
                expect_code = s.get('expect_code')
                if expect_code is not None:
                    expect_code = int(expect_code)
                expect_stdout = s.get('expect_stdout')
                if expect_stdout is not None:
                    expect_stdout = re.compile(expect_stdout)
                expect_stderr = s.get('expect_stderr')
                if expect_stderr is not None:
                    expect_stderr = re.compile(expect_stderr)
                max_output_len = int(s.get('max_output_len', self.DEFAULT_MAX_OUTPUT_LEN))
            except KeyError as e:
                raise SettingError('CommandWatcher not found config key: %s' % e)
            except Exception as e:
                raise SettingError('CommandWatcher settings syntax error: %s' % e)

            if not level:
                raise SettingError('CommandWatcher invalid level: %s' % level_str)
            if all([expect_code is None, expect_stdout is None, expect_stderr is None]):
                raise SettingError('CommandWatcher any of expect_code, expect_stdout or expect_stderr should be set.')

            settings.append((name, level[0], command, expect_code, expect_stdout, expect_stderr, max_output_len))

        super(CommandWatcher, self).__init__(settings=settings)

    def watch(self):
        """
        :return: list of Alert instances
        """
        start_time = datetime.now()

        result = []
        for name, level, command, expect_code, expect_stdout, expect_stderr, max_output_len in self.settings:
            code, stdout, stderr = self._execute_external_command(command)
            if self._should_alert(code, stdout, stderr, expect_code, expect_stdout, expect_stderr):
                message = MSG_CMD_ALERT_FORMAT % {
                    'level': level.get_text(),
                    'name': name,
                    'code': code,
                    'stdout': stdout[:max_output_len],
                    'stderr': stderr[:max_output_len],
                    'expect_code': expect_code,
                    'expect_stdout': expect_stdout.pattern if expect_stdout else None,
                    'expect_stderr': expect_stderr.pattern if expect_stderr else None,
                }
                result.append((level, message))

        if result:
            max_level = max(r[0] for r in result)
            message = MSG_CMD_ALERT % {'server_id': get_server_id(), 'result': '\n'.join(r[1] for r in result)}
            return [Alert(start_time, max_level, MSG_CMD_ALERT_TITLE, message)]
        else:
            return []

    @staticmethod
    def _execute_external_command(command):
        p = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True)
        stdout, stderr = p.communicate()
        return p.returncode, stdout.decode('utf-8'), stderr.decode('utf-8')

    @staticmethod
    def _should_alert(code, stdout, stderr, expect_code, expect_stdout, expect_stderr):
        if expect_code is not None:
            if code != expect_code:
                return True
        if expect_stdout is not None:
            if not expect_stdout.findall(stdout):
                return True
        if expect_stderr is not None:
            if not expect_stderr.findall(stderr):
                return True
        return False



import smtplib
from email.mime.text import MIMEText
from email.header import Header
from email.utils import formatdate
from notifier import Notifier
from easy_alert.i18n import EMAIL_ENCODING
from easy_alert.setting.setting_error import SettingError


class EmailNotifier(Notifier):
    def __init__(self, notify_setting, print_only, logger):
        if not isinstance(notify_setting, dict):
            raise SettingError('EmailNotifier settings not a dict: %s' % notify_setting)
        try:
            super(EmailNotifier, self).__init__(
                notify_setting['group_id'],
                from_address=notify_setting['from_address'],
                to_address_list=notify_setting['to_address_list'],
                smtp_server=notify_setting['smtp_server'],
                smtp_port=notify_setting['smtp_port'],
                print_only=print_only,
                logger=logger)
        except KeyError as e:
            raise SettingError('EmailNotifier not found config key: %s' % e)

    def notify(self, alert):
        subject = self.get_subject(alert)
        body = alert.message

        if self.print_only:
            self.logger.info('Would send a message:')
            self.logger.info('Subject: %s' % subject)
            self.logger.info('From: %s' % self.from_address)
            self.logger.info('To: %s' % self.to_address_list)
            self.logger.info('Body:')
            for line in body.split('\n'):
                self.logger.info(line)
        else:
            msg = self.__create_message(self.from_address, self.to_address_list, subject, body)
            s = smtplib.SMTP(self.smtp_server, self.smtp_port)
            try:
                s.sendmail(self.from_address, self.to_address_list, msg.as_string())
            finally:
                s.close()
            self.logger.info('Sent mail to %s.' % self.to_address_list)

    @staticmethod
    def __create_message(from_addr, to_addr_list, subject, body, encoding=EMAIL_ENCODING):
        """create MIMEText object"""

        msg = MIMEText(body.encode(encoding, 'ignore'), 'plain', encoding)
        msg['Subject'] = Header(subject, charset=encoding, header_name='Subject')
        msg['From'] = from_addr
        msg['To'] = to_addr_list
        msg['Date'] = formatdate(localtime=True)
        return msg



import syslog
from abc import ABCMeta, abstractmethod
from easy_alert.util.case_class import CaseClass


class Logger(CaseClass):
    """abstract logger class"""

    __metaclass__ = ABCMeta

    def __init__(self, ident):
        """
        :param ident: identity for logger
        """
        super(Logger, self).__init__(['ident'])
        self.ident = ident

    def info(self, message):
        self._log_message(syslog.LOG_INFO, message)

    def warn(self, message):
        self._log_message(syslog.LOG_WARNING, message)

    def error(self, message):
        self._log_message(syslog.LOG_ERR, message)

    def traceback(self):
        """do nothing by default"""

    def _log_message(self, priority, message):
        prefix = {
            syslog.LOG_INFO: '[INFO] ',
            syslog.LOG_WARNING: '[WARN] ',
            syslog.LOG_ERR: '[ERROR]'
        }[priority]
        msg = '%s%s' % (prefix, message)
        self._log(priority, msg)

    @abstractmethod
    def _log(self, priority, message):
        """log raw message"""



# -*- coding: utf-8 -*-

EMAIL_ENCODING = 'utf-8'

MSG_DEBUG = u"DEBUG"
MSG_INFO = u"INFO"
MSG_WARN = u"WARN"
MSG_ERROR = u"ERROR"
MSG_CRITICAL = u"CRITICAL"

MSG_PROC_NOT_RUNNING = u'not running'
MSG_PROC_RUNNING = u'%(count)d process(es) are running'
MSG_PROC_STATUS_FORMAT = u'[%(level)s] %(name)s: %(count)s (not "%(condition)s")'
MSG_PROC_ALERT_TITLE = u'Detected Process Abnormality'
MSG_PROC_ALERT = u'Detected the following process abnormality on server %(server_id)r.\n\n%(result)s\n\n=='

MSG_LOG_SUMMARY = u'[%(tag)s]: %(count)d messages'
MSG_LOG_SNIP = u'(snip)'
MSG_LOG_ALERT_TITLE = u'Detected Error Messages'
MSG_LOG_ALERT = u'Detected the following errors on server %(server_id)r.\n\n%(result)s\n=='
MSG_LOG_ALERT_PENDING_TITLE = u'Possible Retention of Log Watcher'
MSG_LOG_ALERT_PENDING = u"""Detected the possible retention of Log Watcher on server %(server_id)r.\n
There exist multiple files which matches the pattern %(pattern)s.\n\n%(paths)s\n
Please check the state of the files, then restart Fluentd or other applications if needed."""

MSG_SSH_STATUS_FORMAT = u'[%(name)s](%(user)s@%(host)s:%(port)d): %(msg)s'
MSG_SSH_ALERT_TITLE = u'Detected SSH Connection Error'
MSG_SSH_ALERT = u'Failed to connect to the following servers using SSH from %(server_id)r.\n\n%(result)s\n\n=='

MSG_CMD_ALERT_FORMAT = u"""[%(level)s] Failed health check: %(name)s
  actual: {code:%(code)d, stdout:%(stdout)s, stderr:%(stderr)s}
  expect: {code:%(expect_code)d, stdout:%(expect_stdout)s, stderr:%(expect_stderr)s}"""
MSG_CMD_ALERT_TITLE = u'Detected Health Check Error'
MSG_CMD_ALERT = u'Detected the following errors on server %(server_id)r.\n\n%(result)s\n\n=='

MSG_HTTP_ALERT_FORMAT = u"""[%(level)s] Failed health check: %(name)s
  url    : %(url)s
  actual : {code:%(code)d, size:%(size)d}
  expect : {code:%(expect_code)d, size:%(expect_size)s, regexp:%(expect_regexp)s}
  message: %(additional_info)s"""
MSG_HTTP_ALERT_TITLE = u'Detected HTTP Connection Error'
MSG_HTTP_ALERT = u'Detected the following errors on server %(server_id)r.\n\n%(result)s\n\n=='

MSG_SUBJECT_FORMAT = u'【%(level)s】[%(group_id)s:%(server_id)s] %(title)s (%(start_time)s)'
